/** 
 * Copyright (c) Krapht, 2011
 * 
 * "LogisticsPipes" is distributed under the terms of the Minecraft Mod Public 
 * License 1.0, or MMPL. Please check the contents of the license located in
 * http://www.mod-buildcraft.com/MMPL-1.0.txt
 */

package net.minecraft.src.buildcraft.krapht;

import java.util.HashMap;
import java.util.LinkedList;

import net.minecraft.src.mod_LogisticsPipes;
import net.minecraft.src.buildcraft.api.Orientations;
import net.minecraft.src.krapht.ItemIdentifier;

public class Router{

	public class LSA {
		public Router source;
		public HashMap<Router, Integer> neighboursWithMetric;
	}
	
	public static LinkedList<Router> allRouters = new LinkedList<Router>();
	
	public HashMap<RoutedPipe, ExitRoute> _adjacent = new HashMap<RoutedPipe, ExitRoute>();
	private final LinkedList<RoutedEntityItem> _outboundItems = new LinkedList<RoutedEntityItem>();
	private final LinkedList<RoutedEntityItem> _inboundItems = new LinkedList<RoutedEntityItem>();

	private static int _LSDVersion = 0;
	private int _lastLSDVersion = 0;
	private final RoutedPipe _pipe;
	
	private static RouteLaser _laser = new RouteLaser();
	
	private static LinkedList<LSA> SharedLSADatabase = new LinkedList<Router.LSA>();
	private LSA _myLsa = new LSA();
		
	/** Map of router -> orientation for all known destinations **/
	private HashMap<Router, Orientations> _routeTable = new HashMap<Router, Orientations>();
	private HashMap<Router, Integer> _routeCosts = new HashMap<Router, Integer>();

	private boolean _blockNeedsUpdate;
	
	public Router(RoutedPipe pipe){
		this._pipe = pipe;
		_myLsa = new LSA();
		_myLsa.source = this;
		_myLsa.neighboursWithMetric = new HashMap<Router, Integer>();
		SharedLSADatabase.add(_myLsa);
		allRouters.add(this);
	}

		
	public void update(boolean doFullRefresh){
		if (doFullRefresh) {
			recheckAdjacent();
			if (_blockNeedsUpdate){
				_pipe.worldObj.markBlockAsNeedsUpdate(this._pipe.xCoord, this._pipe.yCoord, this._pipe.zCoord);	
				_blockNeedsUpdate = false;
			}
			return; //To not update the route table in the same tick as rechecking adjacent
		}
	}
	
	public HashMap<Router, Orientations> getRouteTable(){
		if (_LSDVersion > _lastLSDVersion){
			CreateRouteTable();
			_lastLSDVersion = _LSDVersion;
		}
		return _routeTable;
	}
	
	public RoutedPipe getPipe() {
		return _pipe;
	}
	
	public String getId() {
		return this.toString().substring(this.toString().indexOf('@'));
	}
	

	/**
	 * Rechecks the piped connection to all adjacent routers as well as discover new ones.
	 */
	public void recheckAdjacent()	{
		boolean adjacentChanged = false;
		HashMap<RoutedPipe, ExitRoute> adjacent = PathFinder.getConnectedRoutingPipes(_pipe.container, mod_LogisticsPipes.LOGISTICS_DETECTION_COUNT, mod_LogisticsPipes.LOGISTICS_DETECTION_LENGTH);
		
		for (RoutedPipe pipe : _adjacent.keySet()){
			if(!adjacent.containsKey(pipe))
				adjacentChanged = true;
		}
		
		for (RoutedPipe pipe : adjacent.keySet())	{
			if (!_adjacent.containsKey(pipe)){
				adjacentChanged = true;
				break;
			}
			ExitRoute newExit = adjacent.get(pipe);
			ExitRoute oldExit = _adjacent.get(pipe);
			
			if (newExit.exitOrientation != oldExit.exitOrientation || newExit.metric != oldExit.metric)	{
				adjacentChanged = true;
				break;
			}
		}
		
		if (adjacentChanged){
			_adjacent = adjacent;
			_blockNeedsUpdate = true;
			SendNewLSA();
		}
	}
	
	public void SendNewLSA()
	{
		_myLsa.neighboursWithMetric = new HashMap<Router, Integer>();
		for (RoutedPipe adjacent : _adjacent.keySet()){
			_myLsa.neighboursWithMetric.put(adjacent.router, _adjacent.get(adjacent).metric);
		}
		_LSDVersion++;
		CreateRouteTable();
	}
	

	
	
	/**
	 * Create a route table from the link state database
	 */
	private void CreateRouteTable()	{ 
		//Dijkstra!
		
		/** Map of all "approved" routers and the route to get there **/
		HashMap<Router, LinkedList<Router>> tree =  new HashMap<Router, LinkedList<Router>>();
		/** The cost to get to an "approved" router **/
		HashMap<Router, Integer> treeCost = new HashMap<Router, Integer>();
		
		//Init root(er - lol)
		tree.put(this,  new LinkedList<Router>());
		treeCost.put(this,  0);
		/** The candidate router and which approved router put it in the candidate list **/
		HashMap<Router, Router> candidates = new HashMap<Router, Router>();
		/** The total cost for the candidate route **/
		HashMap<Router, Integer> candidatesCost = new HashMap<Router, Integer>();
		
		//Init candidates
		for (RoutedPipe pipe :  _adjacent.keySet()){
			candidates.put(pipe.router, this);
			candidatesCost.put(pipe.router, _adjacent.get(pipe).metric);
		}

		
		while (!candidates.isEmpty()){
			Router lowestCostCandidateRouter = null;
			int lowestCost = Integer.MAX_VALUE;
			for(Router candidate : candidatesCost.keySet()){
				if (candidatesCost.get(candidate) < lowestCost){
					lowestCostCandidateRouter = candidate;
					lowestCost = candidatesCost.get(candidate);
				}
			}
			
			Router lowestParent = candidates.get(lowestCostCandidateRouter);	//Get the approved parent of the lowest cost candidate
			LinkedList<Router> lowestPath = (LinkedList<Router>) tree.get(lowestParent).clone();	//Get a copy of the route for the approved router 
			lowestPath.addLast(lowestCostCandidateRouter); //Add to the route to get to the candidate
			
			//Approve the candidate
			tree.put(lowestCostCandidateRouter, lowestPath);
			treeCost.put(lowestCostCandidateRouter, lowestCost);
			
			//Remove from candidate list
			candidates.remove(lowestCostCandidateRouter);
			candidatesCost.remove(lowestCostCandidateRouter);
			
			//Add new candidates from the newly approved route
			for (LSA lsa : this.SharedLSADatabase){
				if (lsa.source != lowestCostCandidateRouter) continue;				
				for (Router newCandidate: lsa.neighboursWithMetric.keySet()){
					if (tree.containsKey(newCandidate)) continue;
					int candidateCost = lowestCost + lsa.neighboursWithMetric.get(newCandidate);
					if (candidates.containsKey(newCandidate) && candidatesCost.get(newCandidate) <= candidateCost){
						continue;
					}
					candidates.put(newCandidate, lowestCostCandidateRouter);
					candidatesCost.put(newCandidate, candidateCost);
				}
			}
		}
		
		
		//Build route table
		_routeTable = new HashMap<Router, Orientations>();
		_routeCosts = new HashMap<Router, Integer>();
		for (Router node : tree.keySet())
		{
			LinkedList<Router> route = tree.get(node);
			if (route.size() == 0){
				_routeTable.put(node, Orientations.Unknown);
				continue;
			}
			
			Router firstHop = route.getFirst();
			if (!_adjacent.containsKey(firstHop._pipe)){
				//System.out.println("FirstHop is not adjacent!");
			}
			_routeCosts.put(node, treeCost.get(node));
			_routeTable.put(node, _adjacent.get(firstHop._pipe).exitOrientation);
		}
	}

//	public long getLastSequenceNumber()
//	{
//		return _lastSequenceNumber;
//	}
	
	/**
	 * Flags the last sent LSA as expired. Each router will be responsible of purging it from its database.
	 */
	public void destroy() {
//		if (_lastSentLSA != null){
//			_lastSentLSA.expired = true;
//			//System.out.println("Router removed");
//		}
		if (SharedLSADatabase.contains(_myLsa)){
			SharedLSADatabase.remove(_myLsa);
			_LSDVersion++;
		}
		
		if (allRouters.contains(this)){
			allRouters.remove(this);
		}
			
	}
	
	public LinkedList<Orientations> GetNonRoutedExits()	{
		LinkedList<Orientations> ret = new LinkedList<Orientations>();
		
		outer:
		for (int i = 0 ; i < 6; i++){
			Orientations o = Orientations.values()[i];
			boolean found = false;
			for(ExitRoute route : _adjacent.values()) {
				if (route.exitOrientation == o){
					found = true;
					continue outer; //Its routed
				}
			}
			ret.add(o);	//Its not (might not be a pipe, but its not routed)
		}
		return ret;
	}
	
	public boolean isRoutedExit(Orientations o){
		return !GetNonRoutedExits().contains(o);
	}
	

	public void displayRoutes(){
		_laser.displayRoute(this);
	}
	
	public void displayRouteTo(Router r){
		_laser.displayRoute(this, r);
	}

	
	
	public int getInboundItemsCount(){
		return _inboundItems.size();
	}
	
	public int getOutboundItemsCount(){
		return _outboundItems.size();
	}
	
	public void startTrackingRoutedItem(RoutedEntityItem routedEntityItem){
		if(!_outboundItems.contains(routedEntityItem)){
			_outboundItems.add(routedEntityItem);
		}
	}

	public void startTrackingInboundItem(RoutedEntityItem routedEntityItem){
		if (!_inboundItems.contains(routedEntityItem)){
			_inboundItems.add(routedEntityItem);
		}
	}
	
	public void outboundItemArrived(RoutedEntityItem routedEntityItem){
		if (_outboundItems.contains(routedEntityItem)){
			_outboundItems.remove(routedEntityItem);
		}
	}
	
	public void inboundItemArrived(RoutedEntityItem routedEntityItem){
		if (_inboundItems.contains(routedEntityItem)){
			_inboundItems.remove(routedEntityItem);
		}

		//notify that Item has arrived
		if (getPipe().logic instanceof IRequireReliableTransport){
			((IRequireReliableTransport)getPipe().logic).itemArrived(ItemIdentifier.get(routedEntityItem.item));
		}
	}

	public void itemDropped(RoutedEntityItem routedEntityItem) {
		if (_outboundItems.contains(routedEntityItem)){
			_outboundItems.remove(routedEntityItem);
		}
		if (_inboundItems.contains(routedEntityItem)){
			_inboundItems.remove(routedEntityItem);
		}
	}
}


